#include "Arduino.h"
#include <Wire.h>
#include "FS.h"
#include <SD.h>
#include <SPI.h>
#include <soc/rtc.h>
#include "driver/rtc_io.h"
#include <ESP32Time.h>
#include "esp_sleep.h"
#include <WebServer.h>
#include <WiFi.h>
#include <LittleFS.h>
#include "driver/adc.h"
#include "esp_adc_cal.h"

#include <ESPmDNS.h>

// for PILOT LED
#define PILOT_LED   2
#define POWER_LED   25

// for RTC
#define uS_TO_S_FACTOR 1000000ULL
#define CALIBRATE_ONE(cali_clk) calibrate_one(cali_clk, #cali_clk)
ESP32Time rtc;

// for DHT20
//            DHT20
//          +--------------+
//  VDD ----| 1            |
//  SDA ----| 2  GPIO21    |
//  GND ----| 3  GPIO22    |
//  SCL ----| 4            |
//          +--------------+
//
#define ADR 0x38
#define DHT20_0   17
#define DHT20_1   16
#define DHT20_2    4
#define DHT20_3   15

// SD Card select PIN
#define sd_sck    18  // default
#define sd_mosi   23  // default
#define sd_ss      5  // default
#define sd_miso   19  // default

const char* ssid = "logger_AP";
const char* password = "test1234";
IPAddress local_IP(192, 168, 4, 10);
IPAddress gateway(192, 168, 4, 1);
IPAddress subnet(255, 255, 255, 0);
WebServer server(80);

RTC_DATA_ATTR int   int_time;   //hhmm
RTC_DATA_ATTR int   port_st = 0;
RTC_DATA_ATTR unsigned long  enter_time = 0;   

RTC_DATA_ATTR int   bootCount = 0;   

unsigned long back_time;

void setup(){
  int a;
  String str;
  
    back_time = rtc.getLocalEpoch();
    Serial.begin(115200);

    pinMode(36, INPUT);
    while(!digitalRead(36)) delay(50);
    delay(500);
    // Initialise pilot led
    pinMode(PILOT_LED, OUTPUT);
    digitalWrite(PILOT_LED, HIGH);

    // wakeup_32k
    if(!wakeup_32k()) error_led();
      
    // Initialise DHT20 select port
    pinMode(DHT20_0, OUTPUT);
    pinMode(DHT20_1, OUTPUT);
    pinMode(DHT20_2, OUTPUT);
    pinMode(DHT20_3, OUTPUT);
    digitalWrite(DHT20_0, LOW);
    digitalWrite(DHT20_1, LOW);
    digitalWrite(DHT20_2, LOW);
    digitalWrite(DHT20_3, LOW);
    Wire.begin();

    // Initialise SD 
    SPI.begin(sd_sck, sd_miso, sd_mosi, sd_ss);
    if (!SD.begin(sd_ss)){
        Serial.println("SD initialisation failed!");
        error_led();
    }
    else Serial.println("SD initialisation OK");

    switch(esp_sleep_get_wakeup_cause()){
      case ESP_SLEEP_WAKEUP_TIMER : 
                          Serial.println("ESP_SLEEP_WAKEUP_TIMER"); 
                          bootCount ++;
                          Serial.println("Now: " + String(bootCount));
                          Serial.println("Going to sleep now");
                          measure(true);
                          delay(100);
                          break;
                    
      case ESP_SLEEP_WAKEUP_EXT0  : 
                          Serial.println("ESP_SLEEP_WAKEUP_EXT0");
                          Serial.println("sec: " + String(rtc.getLocalEpoch() - enter_time));
     
      default                     : 
                          Serial.println("ESP_SLEEP_WAKEUP by Reset"); 
                          if(!(int_time & 0x8000)) 
                            rtc.setTime(0, 0, 0, 1, 1, 2025);
                          // Initialise LittleFS
                          if (!LittleFS.begin()) {
                            Serial.println("LittleFS initialisation failed!");
                            error_led();
                          }
                          else Serial.println("LittleFS initialisation OK");
                   
                          // Connect WiFi 
                          WiFi.disconnect();
                          WiFi.softAPdisconnect(true);
                          delay(500);

                          // Connect WiFi AP
                          WiFi.mode(WIFI_AP);
                          WiFi.softAP(ssid, password);
                          WiFi.softAPConfig(local_IP, gateway, subnet);
                          Serial.println(WiFi.softAPIP());
                          delay(500);
  
                          server.on("/", handleRoot);
                          server.onNotFound(handleWebRequests); 
                          server.begin();
                          Serial.println("HTTP server started");
                          break;    
  }
}

void loop(){
  server.handleClient();
}

void handleRoot(){
  String buf,cmd;
  int a, b, c,fl,st;
  File dataFile;
  uint8_t para[10];
  float ht_data[3];
  String week[7] = {"Sun", "Mon", "Tue", "Wed", "Thu", "Fri", "Sat"};

    fl = 1;
    cmd=server.argName(0);
    switch(cmd.toInt()){
      case 1:   // Update & Set Parameter
                Serial.println("Update & Set Parameter");
                get_parameter("1", 3, para);
                port_st = para[0];
                a = int_time & 0x8000;
                int_time = para[1];
                int_time <<= 8;
                int_time |= para[2];
                int_time |= a;
                break;
      
      case 2:   //Set TIME
                dataFile = LittleFS.open("/set_time.html", FILE_READ);
                server.streamFile(dataFile,"text/html");
                dataFile.close();
                fl = 0;
                break;

      case 3:   // Stop
                Serial.println("Set Parameter");
                int_time &= 0x7fff;
                break;
                
      case 4:   // Continue
                Serial.println("Continue");
                b = rtc.getLocalEpoch() - enter_time;
                a = c =  calc_sec();
                a -= b;
                if(a < 1) a = c;
                esp_sleep_enable_timer_wakeup(a * uS_TO_S_FACTOR);
                esp_sleep_enable_ext0_wakeup(GPIO_NUM_36,0);
                enter_time = rtc.getLocalEpoch();
                esp_deep_sleep_start();
                delay(100);
                break;
                
      case 5:   // Start
                dataFile = SD.open("/data0.dat", FILE_WRITE);
                dataFile.close();
                dataFile = SD.open("/data1.dat", FILE_WRITE);
                dataFile.close();
                bootCount = 0;

                server.send(200, "text/plain", "");
                Serial.println("Now: " + String(bootCount));
                Serial.println("Going to sleep now");
                int_time |= 0x8000;
                measure(false);
                delay(100);
                break;
                
      case 10:  //Set TIME 
                get_parameter("10", 5, para);
                rtc.setTime(0, para[4], para[3], para[2], para[1], 2000 + para[0]);
                break;
                
      case 80:  // Send Logger parameter
                buf = "Idle, ,";
                if(int_time & 0x8000){
                  a = calc_sec();
                  b = rtc.getLocalEpoch() - enter_time;
                  Serial.println("a: " + String(a) + " b: " + String(b));
                  a -= b;
                  buf = "Running," + String(a) + "(sec),";
                }
                // baterry data
                buf += (String(get_vol() / 1000.0, 2) + "v,");
                
                // time data
                buf += (String(rtc.getYear()) + " ");
                buf += (String(rtc.getMonth() + 1) + "/" + String(rtc.getDay()) + " ");
                buf += (week[rtc.getDayofWeek()] + " ");
                buf += (rtc.getTime() + ",");
          
                // temp & hum
                b = 1;
                for(a = 0; a < 4; a ++){
                  if(b & port_st){
                    if(get_t_h(a, ht_data)){
                      buf += (String(ht_data[1],1) + "'C / ");
                      buf += (String(ht_data[0],1) + "%,");
                    }
                    else
                      buf += "Can't measure,";
                  }
                  else
                    buf += "No use,";
                  b <<= 1;  
                }
                
                // check buttom
                buf += (String(port_st) + ",");
                
                // Interval hour & minute
                buf += (String((int_time >> 8) & 0x7f) + ",");
                buf += (String(int_time & 0xff) + ",");
                
                server.send(200, "text/plain", buf);
                
                fl=0;
                break;
                
      case 90:  // Send Timer parameter
                buf = (String(rtc.getYear() - 2000) + ",");
                buf += (String(rtc.getMonth() + 1) + ",");
                buf += (String(rtc.getDay()) + ",");
                buf += (String(rtc.getHour(true)) + ",");
                buf += (String(rtc.getMinute()) + ",");

                Serial.println(buf);
                
                server.send(200, "text/plain", buf);
                fl=0;
                break;
                
      case 99:  // Quite
                break;      
    }
    if(fl) {
        dataFile = LittleFS.open("/logger_esp32.html", FILE_READ);
        server.streamFile(dataFile,"text/html");
        dataFile.close();
    }
    
}

void measure(bool flg){
  String buf;
  float ht_data[2];
  File dataFile;
  int a, b;
                
    buf = (String(bootCount) + ","); 
    buf += (String(get_vol() / 1000.0, 2) + ",");
    buf += (rtc.getTimeDate() + ",");

    b = 1;
    for(a = 0; a < 4; a ++){
      buf += (String(a + 1) + ":,");
      if(port_st & b){
        if(get_t_h(a, ht_data)){
          buf += (String(ht_data[1],1) + ",");
          buf += (String(ht_data[0],1) + ",");
        }
        else buf += "Not,Not,";
      }
      else buf += "No use,";
      b <<= 1;  
    }
    
    dataFile = SD.open("/data0.dat", FILE_APPEND);
    dataFile.println(buf);
    dataFile.close();
    
    dataFile = SD.open("/data1.dat", FILE_APPEND);
    dataFile.println(buf);
    dataFile.close();
    
    a = calc_sec();
    if(flg) a -= (rtc.getLocalEpoch() - back_time);
                
    esp_sleep_enable_timer_wakeup(a * uS_TO_S_FACTOR);
    esp_sleep_enable_ext0_wakeup(GPIO_NUM_36,0);
    enter_time = rtc.getLocalEpoch();
    esp_deep_sleep_start();
}

int calc_sec(){
  int a,b;
    a = (int_time >> 8) &  0x7f;
    b = int_time & 0xff;
    a *= 3600;
    b *= 60;
    a += b;
    return(a);
}

bool get_t_h(int p_no, float* ht_buf){
  uint8_t buf[8];
  long a;
  int flg;
  int dht_port[4] = {17, 16, 4, 15};
  bool flg_ok;

    for(a = 0; a < 4; a ++) digitalWrite(dht_port[a], LOW);
    digitalWrite(dht_port[p_no], HIGH);
    delay(100);
    flg = 1;
    while (flg)
    {
        Wire.beginTransmission(ADR);
        Wire.write(0xac);
        Wire.write(0x33);
        Wire.write(0x00);
        Wire.endTransmission();
        delay(100);

        Wire.requestFrom(ADR, 6);
        for (uint8_t i = 0; i < 6; i++)  buf[i] = Wire.read();

        if (buf[0] & 0x80){
          delay(100);
          flg ++;
          if(flg > 10) break;
        }
        else flg = 0;
    }

    flg_ok = true;
    if(flg){
      Serial.println("Measurement not Comp");
      flg_ok = false;
    }
    else{
      a = buf[1] << 8;
      a |= buf[2];
      a <<= 4;
      a |= ((buf[3] >> 4) & 0x0f);
      ht_buf[0] = a / 10485.76;            //hum

      a = (buf[3] & 0xf);
      a <<= 8;
      a |= buf[4];
      a <<= 8;
      a |= buf[5];
      ht_buf[1] = a / 5242.88 - 50;        // temp

    }
    
    digitalWrite(dht_port[p_no], LOW);
    return(flg_ok);
}

uint32_t get_vol(){
  uint32_t  vol, vol_sum;  
  esp_adc_cal_characteristics_t adcChar;
  int a;
  
  // Initialise adc 
    adc1_config_width(ADC_WIDTH_BIT_12);
    adc1_config_channel_atten(ADC1_CHANNEL_3, ADC_ATTEN_DB_11);   // ADC1_CHANNEL_3 -> GPIO39
    esp_adc_cal_characterize(ADC_UNIT_1, ADC_ATTEN_DB_11, ADC_WIDTH_BIT_12, 1100, &adcChar);

    vol_sum = 0;
    for(a = 0; a < 10; a++){
      esp_adc_cal_get_voltage(ADC_CHANNEL_3, &adcChar, &vol);
      vol_sum += vol;
      delay(50);
    }

    vol = vol_sum / 10 * 80 / 50;
    return(vol);
}

void get_parameter(char* pt, int num, uint8_t* para){
  int a, b, c;
  String buf, temp;
    
    buf = server.arg(pt);
    b = c = 0;
    for(a = 0; a < num; a ++){
      while(buf[b] != ',') b ++;
      temp = buf.substring(c, b --);
      para[a] = temp.toInt();
      b += 2;    
      c = b;
    }
}

void error_led(){
  while(1){
    digitalWrite(PILOT_LED, LOW);
    delay(500);
    digitalWrite(PILOT_LED, HIGH);
    delay(500);
  }
}

void handleWebRequests(){
  String dataType = "text/plain";
  String path;
  File dataFile;
  int a;

    a = 1;
    path = server.uri();
    if(path.endsWith(".txt")) dataType = "text/plain";
    else if(path.endsWith(".css")) dataType = "text/css";
    else if(path.endsWith(".js")) dataType = "application/javascript";
    else if(path.endsWith(".png")) dataType = "image/png";
    else if(path.endsWith(".html")) dataType = "text/html";
    else if(path.endsWith(".jpg")) dataType = "image/jpeg";
    else if(path.endsWith(".dat")){
      dataType = "text/plain";
      a = 0;
    }
    delay(5);

    if(a) dataFile = LittleFS.open(path.c_str(), "r");
    else dataFile = SD.open(path.c_str(), "r");
    server.streamFile(dataFile, dataType);
    dataFile.close();
    delay(5);
}

int wakeup_reason()
{
  esp_sleep_wakeup_cause_t wakeup_reason;

  wakeup_reason = esp_sleep_get_wakeup_cause();
  if(wakeup_reason == ESP_SLEEP_WAKEUP_TIMER) Serial.println("Wakeup caused by timer");
  else Serial.printf("Wakeup was not caused by deep sleep: %d\n", wakeup_reason);

  print_slow_clock_source();

  rtc_clk_32k_bootstrap(10);
  rtc_clk_slow_freq_set(RTC_SLOW_FREQ_32K_XTAL);
  rtc_clk_32k_enable(true);

  uint32_t cal_32k = CALIBRATE_ONE(RTC_CAL_32K_XTAL);

  if (cal_32k == 0) Serial.println("32K XTAL OSC has not started up");
  else Serial.println("32K XTAL OSC Calibrated");

  print_slow_clock_source();
  return(wakeup_reason);
}

bool wakeup_32k(){
  rtc_clk_32k_bootstrap(10);
  rtc_clk_slow_freq_set(RTC_SLOW_FREQ_32K_XTAL);
  rtc_clk_32k_enable(true);

  uint32_t cal_32k = CALIBRATE_ONE(RTC_CAL_32K_XTAL);

  if (cal_32k == 0){
    Serial.println("32K XTAL OSC has not started up");
    return(false);
  }
  else {
    Serial.println("32K XTAL OSC Calibrated");
    return(true);
  }
}

void print_slow_clock_source()
{
  rtc_slow_freq_t slow_clk = rtc_clk_slow_freq_get();
  Serial.print("Slow clk source: ");
  switch(slow_clk)
  {
    case RTC_SLOW_FREQ_RTC: Serial.println("Internal 150kHz"); break;
    case RTC_SLOW_FREQ_32K_XTAL: Serial.println("External 32kHz"); break;
    case RTC_SLOW_FREQ_8MD256: Serial.println("Internal 8MHz");
  }
}

const float factor = (1 << 19) * 1000.0f;
static uint32_t calibrate_one(rtc_cal_sel_t cal_clk, const char *name)
{
    const uint32_t cal_count = 1000;
    uint32_t cali_val;

    printf("%s:\n", name);
    for (int i = 0; i < 10; ++i)
    {
        printf("calibrate (%d): ", i);
        cali_val = rtc_clk_cal(cal_clk, cal_count);
        printf("%.3f kHz\n", factor / (float)cali_val);
    }
    return cali_val;
}


//  -- END OF FILE --
